/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <efl_extension.h>
#include "$(appName).h"
#include "view.h"
#include "view_defines.h"

#define IMAGE_FRAME "images/frame.png"
#define FRAME_BORDER 3
#define OBJECT_BORDER 3
#define OBJECT_DATA "object"

typedef enum {CL_PEN = 0, CL_FILL, CL_BG} color_layer_t;
typedef enum {CLEAN = 0, FREEHAND, COLOR_SELECTOR, LINE, CIRCLE, RECTANGLE, SELECT, REMOVE } mode_type_t;

typedef struct {
	mode_type_t type;
	Eina_List *parts;
} object_t;

typedef struct _color {
	color_layer_t layer;
	int r;
	int g;
	int b;
	int a;
} color_t;

static struct view_info {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *layout;
	Evas_Object *draw_area;
	Evas_Object *color_selector_panel;
	object_t *current_object;
	object_t *selected;
	Evas_Object *selection_frame;
	Evas *evas;
	color_t cl_pen;
	color_t cl_fill;
	color_t cl_bg;
	/* the color of currently selected object's layer:
	 * pen, fill, background. */
	color_t *color;
	Evas_Coord_Point start;
	Evas_Coord_Point prev;
	Evas_Coord_Point curr;
	mode_type_t mode;
	mode_type_t prev_mode;
	bool mouse_pressed;
	Eina_List *objects;
	int win_layer;
} s_info = {
	.win = NULL,
	.conform = NULL,
	.layout = NULL,
	.draw_area = NULL,
	.color_selector_panel = NULL,
	.evas = NULL,
	.current_object = NULL,
	.selected = NULL,
	.selection_frame = NULL,
	.cl_pen = { .layer = CL_PEN, .r = 0, .g = 0, .b = 0, .a = 255, },
	.cl_fill = { .layer = CL_FILL, .r = 255, .g = 255, .b = 255, .a = 255, },
	.cl_bg = { .layer = CL_BG, .r = 255, .g = 255, .b = 255, .a = 255, },
	.color = NULL,
	.start = {0, 0},
	.prev = {0, 0},
	.curr = {0, 0},
	.mode = SELECT,
	.mouse_pressed = false,
	.objects = NULL,
	.win_layer = 0,
};

static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info);
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info);
static void _mouse_down_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void _mouse_up_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void _mouse_move_cb(void *data, Evas *e, Evas_Object *obj, void *event_info);
static void _toolbar_button_click_cb(void *data, Evas_Object *obj, const char *emission, const char *source);
static void _palette_tab_click_cb(void *data, Evas_Object *obj, const char *emission, const char *source);
static void _edje_message_cb(void *data, Evas_Object *obj, Edje_Message_Type type, int id, void *msg);
static void _object_clicked_cb(void *data, Evas *e, Evas_Object *selected, void *event_info);
static void _get_app_resource(const char *edj_file_in, char *edj_path_out);
static Eina_Bool _create_main_layout(void);
static Eina_Bool _create_color_selector_panel(void);
static Eina_Bool _create_selection_frame(void);
static Eina_Bool _update_selection_frame(void);
static Eina_Bool _create_draw_area(void);
static int _get_draw_area_top(void);
static void _clean_draw_area(void);
static void _remove_selected_object(void);
static void _toggle_color_selector(void);
static void _apply_color(color_t *color);
static void _send_color_message(Evas_Object *obj, int msg_id, color_t *color);
static void _select_item(object_t *item);
static void _unselect_item(void);
static void _update_item(void);
static void _update_freehand_item(void);
static void _update_rect_item(void);
static void _update_circle_item(void);
static void _update_line_item(void);
static void _move_item(void);
static object_t *_create_object(mode_type_t type);
static void _add_object_part(object_t *object, Evas_Object *part);
static void _create_rect(void);
static Evas_Object *_add_rect_part(object_t *object, int r, int g, int b, int a, int x, int y);
static void _create_line(void);
static Evas_Object *_create_circle_part(object_t *object, int r, int g, int b, int a);
static void _update_circle_part(Evas_Object *part, int radius_x, int radius_y, int center_x, int center_y);
static void _get_item_geometry(object_t *item, int *min_x, int *min_y, int *max_x, int *max_y);
static Evas_Object *_get_first_part_of_current_object(void);
static Eina_Bool _is_cursor_in_selection_frame(void);
static void _set_toolbar_mode(mode_type_t mode);
static bool _is_same_string(const char *s1, const char *s2);

/**
 * @brief Creates essential objects: window, conformant and layout.
 * @param[in] user_data The user data.
 * @return This function returns EINA_TRUE if the view was created successfully,
 * otherwise EINA_FALSE is returned.
 */
Eina_Bool view_create(void *user_data)
{
	s_info.color = &s_info.cl_pen;

	/* Create the window */
	s_info.win = view_create_win(PACKAGE);
	if (s_info.win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a window.");
		return EINA_FALSE;
	}

	s_info.win_layer = evas_object_layer_get(s_info.win);

	/* Create the conformant */
	s_info.conform = view_create_conformant_without_indicator(s_info.win);
	if (s_info.conform == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a conformant");
		return EINA_FALSE;
	}

	if (!_create_main_layout())
		return EINA_FALSE;

	/* Show the window after main view is set up */
	evas_object_show(s_info.win);

	return EINA_TRUE;
}

/**
 * @brief Creates a basic window named package.
 * @param[in] pkg_name Name of the window
 */
Evas_Object *view_create_win(const char *pkg_name)
{
	Evas_Object *win = NULL;

	/*
	 * Window
	 * Create and initialize elm_win.
	 * elm_win is mandatory to manipulate the window.
	 */
	win = elm_win_util_standard_add(pkg_name, pkg_name);
	elm_win_conformant_set(win, EINA_TRUE);
	elm_win_autodel_set(win, EINA_TRUE);
	elm_win_indicator_mode_set(win, ELM_WIN_INDICATOR_HIDE);
	elm_win_indicator_opacity_set(win, ELM_WIN_INDICATOR_OPAQUE);

	evas_object_smart_callback_add(win, "delete,request", _delete_win_request_cb, NULL);

	return win;
}

/**
 * @brief Creates a layout object for parent object based on provided EDJE script.
 * @param[in] parent The parent object for layout object.
 * @param[in] edj_file_name The relative path to the layout EDJE script.
 * @param[in] edj_group The name of the group to be loaded from the EDJE script.
 * @param[in] target_part_name The EDJE part's name where the layout is to be set.
 * @return The function returns layout object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_layout(Evas_Object *parent, const char *edj_file_name, const char *edj_group, const char *target_part_name)
{
	char edj_path[PATH_MAX] = {0, };
	Evas_Object *layout = NULL;

	_get_app_resource(edj_file_name, edj_path);

	layout = elm_layout_add(parent);
	if (!layout) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_add() failed.");
		return NULL;
	}

	if (!elm_layout_file_set(layout, edj_path, edj_group)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_file_set() failed.");
		return NULL;
	}

	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	if (target_part_name)
		elm_object_part_content_set(parent, target_part_name, layout);

	return layout;
}

/**
 * @brief Creates a conformant without indicator for wearable app.
 * @param[in] win The object to which you want to set this conformant
 * Conformant is mandatory for base GUI to have proper size
 */
Evas_Object *view_create_conformant_without_indicator(Evas_Object *win)
{
	/*
	 * Conformant
	 * Create and initialize elm_conformant.
	 * elm_conformant is mandatory for base GUI to have proper size
	 * when indicator or virtual keypad is visible.
	 */
	Evas_Object *conform = NULL;

	if (win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "window is NULL.");
		return NULL;
	}

	conform = elm_conformant_add(win);
	evas_object_size_hint_weight_set(conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(win, conform);

	evas_object_show(conform);

	return conform;
}

/**
 * @brief Destroys window and frees its resources.
 */
void view_destroy(void)
{
	if (s_info.win == NULL)
		return;

	evas_object_del(s_info.win);
}

/**
 * @brief Internal callback function invoked when the main window needs to be destroyed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

/**
 * @brief Internal callback function invoked on HW Back button press.
 * @param[in] data The user data passed to the eext_object_event_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	if (evas_object_visible_get(s_info.color_selector_panel))
		_toggle_color_selector();
	else
		elm_win_lower(s_info.win);

	return;
}

/**
 * @brief Internal callback function invoked on mouse down event.
 * This function is responsible for point coordinates storage where the
 * mouse button was clicked (tap).
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] e The evas surface of the clicked object.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information about this event.
 */
static void _mouse_down_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	Evas_Event_Mouse_Down *mouse = (Evas_Event_Mouse_Down *)event_info;

	s_info.mouse_pressed = true;

	s_info.start.x = mouse->canvas.x;
	s_info.start.y = mouse->canvas.y;
	s_info.prev.x = mouse->canvas.x;
	s_info.prev.y = mouse->canvas.y;
	s_info.curr.x = mouse->canvas.x;
	s_info.curr.y = mouse->canvas.y;

	if (s_info.mode != SELECT)
		_unselect_item();
}

/**
 * @brief Internal callback function invoked on mouse up event.
 * This function selects an object under the mouse pointer.
  * @param[in] data The user data passed to the callback attachment function.
 * @param[in] e The evas surface of the clicked object.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information about this event.
 */
static void _mouse_up_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	s_info.mouse_pressed = false;
	_select_item(s_info.current_object);
	s_info.current_object = NULL;
}

/**
 * @brief Internal callback function invoked on mouse move event.
 * This function is responsible tracking the mouse pointer movement path
 * and finally updates existing or drawn object.
 * Mouse pointer movement tracking is performed to avoid moving/drawing
 * objects outside the drawing area.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] e The evas surface of the clicked object.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information about this event.
 */
static void _mouse_move_cb(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	int x = 0;
	int y = 0;
	int w = 0;
	int h = 0;
	Evas_Event_Mouse_Move *mouse = (Evas_Event_Mouse_Move *)event_info;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	s_info.prev.x = s_info.curr.x;
	s_info.prev.y = s_info.curr.y;

	if (s_info.curr.y < _get_draw_area_top())
		return;

	if (mouse->cur.canvas.x < x)
		s_info.curr.x = x;
	else if (mouse->cur.canvas.x > x + w)
		s_info.curr.x = x + w;
	else
		s_info.curr.x = mouse->cur.canvas.x;

	if (mouse->cur.canvas.y < y)
		s_info.curr.y = y;
	else if (mouse->cur.canvas.y > y + h)
		s_info.curr.y = y + h;
	else
		s_info.curr.y = mouse->cur.canvas.y;

	_update_item();
}

/**
 * @brief Internal callback function invoked on toolbar item click.
 * This callback function responds to the signal emitted from the EDJE script.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The layout object emitting the signal.
 * @param[in] emission The name of the signal emitted.
 * @param[in] source The name of the part emitting the signal.
 */
static void _toolbar_button_click_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	mode_type_t mode;

	if (_is_same_string(source, PART_TOOLBAR_BUTTON"1"))
		mode = CLEAN;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"2"))
		mode = FREEHAND;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"3"))
		mode = COLOR_SELECTOR;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"4"))
		mode = LINE;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"5"))
		mode = CIRCLE;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"6"))
		mode = RECTANGLE;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"7"))
		mode = SELECT;
	else if (_is_same_string(source, PART_TOOLBAR_BUTTON"8"))
		mode = REMOVE;
	else
		return;

	switch (mode) {
	case CLEAN:
		_clean_draw_area();
		break;
	case COLOR_SELECTOR:
		_toggle_color_selector();
		break;
	case REMOVE:
		_remove_selected_object();
		break;
	default:
		_set_toolbar_mode(mode);
	}
}

/**
 * @brief Internal callback function invoked on palette tab click.
 * This callback function responds to the signal emitted from the EDJE script.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] obj The layout object emitting the signal.
 * @param[in] emission The name of the signal emitted.
 * @param[in] source The name of the part emitting the signal.
 */
static void _palette_tab_click_cb(void *data, Evas_Object *obj, const char *emission, const char *source)
{
	if (_is_same_string(source, PART_PALETTE_TAB_PEN))
		s_info.color = &s_info.cl_pen;
	else if (_is_same_string(source, PART_PALETTE_TAB_FILL))
		s_info.color = &s_info.cl_fill;
	else if (_is_same_string(source, PART_PALETTE_TAB_BACKGROUND))
		s_info.color = &s_info.cl_bg;
	else
		return;

	/* The message below performs color selection in the color palette. */
	_send_color_message(s_info.color_selector_panel, MSG_ID_COLOR_SELECT, s_info.color);
}

/**
 * @brief Internal function which is responsible for received EDJE message processing.
 * If the message of type EDJE_MESSAGE_INT_SET with identifier MSG_ID_COLOR_SELECTED,
 * containing 5 arguments is received then new drawing color is applied.
 * @param[in] data The user data passed to the attachment function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] type The type of the message received.
 * @param[in] id The message identifier.
 * @param[in] msg The received message.
 */
static void _edje_message_cb(void *data, Evas_Object *obj, Edje_Message_Type type, int id, void *msg)
{
	Edje_Message_Int_Set *message = (Edje_Message_Int_Set *)msg;

	if (type != EDJE_MESSAGE_INT_SET) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_edje_message_cb() failed. Unhandled message type received.");
		return;
	} else if (id != MSG_ID_COLOR_SELECTED) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_edje_message_cb() failed. Unhandled message id received.");
		return;
	} else if (message->count != 4) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_edje_message_cb() failed. Unhandled number of arguments received.");
		return;
	}

	if (!s_info.color) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_edje_message_cb() failed. The object layer is not selected for color application.");
		return;
	}

	s_info.color->r = message->val[0];
	s_info.color->g = message->val[1];
	s_info.color->b = message->val[2];
	s_info.color->a = message->val[3];

	_apply_color(s_info.color);
}

/**
 * @brief Internal callback function invoked on mouse down event over existing
 * object (already drawn).
 * This function is responsible for existing (already drawn) object selection
 * and selection frame display. If an object is already selected then it
 * becomes unselected.
 * @param[in] data The user data passed to the callback attachment function.
 * @param[in] e The evas surface of the clicked object.
 * @param[in] selected The object invoking this callback function.
 * @param[in] event_info The structure containing the information about this event.
 */
static void _object_clicked_cb(void *data, Evas *e, Evas_Object *selected, void *event_info)
{
	Eina_Bool frame_visibility = EINA_FALSE;

	if (!s_info.selection_frame)
		return;

	frame_visibility = evas_object_visible_get(s_info.selection_frame);
	evas_object_hide(s_info.selection_frame);

	if (frame_visibility)
		evas_object_show(s_info.selection_frame);

	if (selected == s_info.selection_frame)
		return;

	if (selected == s_info.win || selected == s_info.layout) {
		if (!_is_cursor_in_selection_frame())
			_unselect_item();

		return;
	}

	_select_item(evas_object_data_get(selected, OBJECT_DATA));
}

/**
 * @brief Internal function which creates fully qualified path to the provided resource file.
 * @param[in] edj_file_in The file name.
 * @param[out] edj_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_app_resource(const char *edj_file_in, char *edj_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, PATH_MAX, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates the main layout and all its descendants.
 * @return The function returns 'EINA_TRUE' if the main layout was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_main_layout(void)
{
	Evas_Object *toolbar_layout;

	s_info.layout = view_create_layout(s_info.conform, EDJ_MAIN, GRP_MAIN, "default");
	if (!s_info.layout)
		return EINA_FALSE;

	eext_object_event_callback_add(s_info.layout, EEXT_CALLBACK_BACK, _layout_back_cb, NULL);

	toolbar_layout = view_create_layout(s_info.layout, EDJ_TOOLBAR, GRP_TOOLBAR, PART_MAIN_TOOLBAR);
	if (!toolbar_layout)
		return EINA_FALSE;

	elm_layout_signal_callback_add(toolbar_layout, SIGNAL_TOOLBAR_BUTTON_CLICKED, PART_TOOLBAR_BUTTON"*", _toolbar_button_click_cb, NULL);

	if (!_create_draw_area())
		return EINA_FALSE;

	if (!_create_color_selector_panel())
		return EINA_FALSE;

	return _create_selection_frame();
}

/**
 * @brief Internal function which creates the color selector panel.
 * @return The function returns 'EINA_TRUE' if the color selector panel
 * was created successfully, otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_color_selector_panel(void)
{
	Evas_Object *layout;

	layout = view_create_layout(s_info.layout, EDJ_PALETTE, GRP_PALETTE, PART_MAIN_PALETTE);
	if (!layout)
		return EINA_FALSE;

	evas_object_layer_set(layout, EVAS_LAYER_MAX);

	edje_object_message_handler_set(elm_layout_edje_get(layout), _edje_message_cb, NULL);
	elm_layout_signal_callback_add(layout, SIGNAL_PALETTE_TAB_CLICKED, PART_PALETTE_TABS, _palette_tab_click_cb, NULL);

	s_info.color_selector_panel = layout;

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates the selection frame.
 * @return The function returns 'EINA_TRUE' if the selection frame
 * was created successfully, otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_selection_frame(void)
{
	char path[PATH_MAX] = {0,};

	s_info.selection_frame = evas_object_image_filled_add(s_info.evas);
	if (!s_info.selection_frame) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_image_filled_add() failed.");
		return EINA_FALSE;
	}

	_get_app_resource(IMAGE_FRAME, path);

	evas_object_image_file_set(s_info.selection_frame, path, NULL);
	evas_object_image_border_set(s_info.selection_frame, FRAME_BORDER, FRAME_BORDER, FRAME_BORDER, FRAME_BORDER);
	evas_object_layer_set(s_info.selection_frame, EVAS_LAYER_MAX - 1);
	evas_object_repeat_events_set(s_info.selection_frame, EINA_TRUE);

	return EINA_TRUE;
}

/**
 * @brief Internal function which updates the selection frame (size and position).
 * @return The function returns 'EINA_TRUE' if the selection frame
 * was updated successfully, otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _update_selection_frame(void)
{
	int min_x = 0;
	int min_y = 0;
	int max_x = 0;
	int max_y = 0;

	if (!s_info.selected) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_update_selection_frame() failed. None item is selected.");
		return EINA_FALSE;
	}

	_get_item_geometry(s_info.selected, &min_x, &min_y, &max_x, &max_y);

	evas_object_move(s_info.selection_frame, min_x - 1, min_y - 1);
	evas_object_resize(s_info.selection_frame, max_x - min_x + 1, max_y - min_y + 1);
	evas_object_show(s_info.selection_frame);

	return EINA_TRUE;
}

/**
 * @brief Internal function which creates drawing area and attaches mouse callbacks
 * for drawing functionality.
 * @return This function returns 'EINA_TRUE' if the drawing area was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _create_draw_area(void)
{
	s_info.draw_area = view_create_layout(s_info.layout, EDJ_DRAW_AREA, GRP_DRAW_AREA, PART_MAIN_DRAW_AREA);
	if (!s_info.draw_area)
		return EINA_FALSE;

	s_info.evas = evas_object_evas_get(s_info.win);
	if (!s_info.evas) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_evas_get() failed.");
		return EINA_FALSE;
	}

	evas_object_event_callback_add(s_info.draw_area, EVAS_CALLBACK_MOUSE_DOWN, _mouse_down_cb, NULL);
	evas_object_event_callback_add(s_info.draw_area, EVAS_CALLBACK_MOUSE_UP, _mouse_up_cb, NULL);
	evas_object_event_callback_add(s_info.draw_area, EVAS_CALLBACK_MOUSE_MOVE, _mouse_move_cb, NULL);

	return EINA_TRUE;
}

/**
 * @brief Internal function which gets the top coordinate of the drawing area.
 * This function takes into account the visibility state if the color palette panel.
 * @return This function returns the y coordinate of the upper bound of the
 * drawing area.
 */
static int _get_draw_area_top(void)
{
	int y = 0;
	int h = 0;
	int y_top = 0;

	if (evas_object_visible_get(s_info.color_selector_panel)) {
		evas_object_geometry_get(s_info.color_selector_panel, NULL, &y, NULL, &h);
		y_top = y + h;
	} else {
		evas_object_geometry_get(s_info.draw_area, NULL, &y_top, NULL, NULL);
	}

	return y_top;
}
/**
 * @brief Internal function which cleans the entire drawing area by removing
 * all the objects already drawn.
 */
static void _clean_draw_area(void)
{
	Eina_List *l = NULL;
	object_t *obj = NULL;
	Evas_Object *part = NULL;
	Eina_List *part_list = NULL;

	EINA_LIST_FOREACH(s_info.objects, l, obj) {
		if (!obj)
			continue;

		EINA_LIST_FOREACH(obj->parts, part_list, part) {
			if (!part)
				continue;

			evas_object_del(part);
		}

		eina_list_free(obj->parts);
		free(obj);
	}

	s_info.objects = eina_list_free(s_info.objects);
	evas_object_hide(s_info.selection_frame);
}

/**
 * @brief Internal function which removes currently selected object
 * from the drawing area.
 */
static void _remove_selected_object(void)
{
	Eina_List *l = NULL;
	Evas_Object *part = NULL;

	if (!s_info.selected)
		return;

	EINA_LIST_FOREACH(s_info.selected->parts, l, part) {
		if (!part)
			continue;

		evas_object_del(part);
	}

	s_info.objects = eina_list_remove(s_info.objects, s_info.selected);
	free(s_info.selected);
	_unselect_item();
}

/**
 * @brief Internal function which toggles the color selector panel visibility.
 * Once the color selection panel becomes visible, the appropriate color cells
 * are checked to denote current user's choice of colors.
 */
static void _toggle_color_selector(void)
{
	Edje_Message_Int msg = {0,};
	bool is_visible = !evas_object_visible_get(s_info.color_selector_panel);

	msg.val = (int)is_visible;

	edje_object_message_send(elm_layout_edje_get(s_info.layout), EDJE_MESSAGE_INT, MSG_ID_PALETTE, (void *)&msg);

	if (is_visible)
		/* The message below performs color selection in the color palette. */
		_send_color_message(s_info.color_selector_panel, MSG_ID_COLOR_SELECT, s_info.color);
}

/**
 * @brief Internal function which applies changed color to the currently selected object
 * or to the entire background.
 * @param[in] color The color to be applied.
 */
static void _apply_color(color_t *color)
{
	Eina_List *l = NULL;
	Evas_Object *part = NULL;

	if (color->layer == CL_BG)
		_send_color_message(s_info.draw_area, MSG_ID_COLOR_BG_SET, s_info.color);

	if (!s_info.selected)
		return;

	if (s_info.selected->type == LINE || s_info.selected->type == FREEHAND) {
		if (color->layer == CL_PEN)
			EINA_LIST_FOREACH(s_info.selected->parts, l, part) {
				if (!part)
					continue;

				evas_object_color_set(part, color->r, color->g, color->b, color->a);
			}
	} else {
		if (color->layer == CL_PEN)
			part = eina_list_nth(s_info.selected->parts, 0);
		else if (color->layer == CL_FILL)
			part = eina_list_nth(s_info.selected->parts, 1);

		evas_object_color_set(part, color->r, color->g, color->b, color->a);
	}
}

/**
 * @brief Internal function which sends a message containing color specification.
 * This message may be sent with the following identifiers:
 * - MSG_ID_COLOR_SELECT - for color selection from the color palette;
 * - MSG_ID_COLOR_BG_SET - for drawing area background color change.
 * @param[in] obj The target object for the message.
 * @param[in] msg_id The message identifier. The following ids are supported
 * by the EDJE script: MSG_ID_COLOR_SELECT, MSG_ID_COLOR_BG_SET
 * @param[in] color The color to be applied.
 */
static void _send_color_message(Evas_Object *obj, int msg_id, color_t *color)
{
	Edje_Message_Int_Set *msg = (Edje_Message_Int_Set *)malloc(sizeof(Edje_Message_Int_Set) + 3 * sizeof(int));
	if (!msg) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_send_color_message() failed. Memory allocation error.");
		return;
	}

	msg->count = 4;
	msg->val[0] = color->r;
	msg->val[1] = color->g;
	msg->val[2] = color->b;
	msg->val[3] = color->a;

	edje_object_message_send(elm_layout_edje_get(obj), EDJE_MESSAGE_INT_SET, msg_id, (void *)msg);
}

/**
 * @brief Internal function which performs provided object selection.
 * This function is responsible for object's reference storage
 * and selection frame display.
 * @param[in] item The object to be selected and stored.
 */
static void _select_item(object_t *item)
{
	if (!item) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_select_item() failed.");
		return;
	}

	s_info.selected = item;
	_update_selection_frame();
}

/**
 * @brief Internal function which unselects already selected object.
 * This function is responsible for selected object's reference cleanup
 * and selection frame removal.
 */
static void _unselect_item(void)
{
	s_info.selected = NULL;
	evas_object_hide(s_info.selection_frame);
}

/**
 * @brief Internal function responsible for object update.
 * The object update relies on drawing an object or its reposition.
 * The operation to be performed depends on currently selected operational mode
 * toggled by the user form application's toolbar. Supported operations are:
 * selected object reposition and new objects drawing: freehand, rectangle,
 * ellipse and line.
 */
static void _update_item(void)
{
	switch (s_info.mode) {
	case FREEHAND:
		_update_freehand_item();
		break;
	case RECTANGLE:
		_update_rect_item();
		break;
	case CIRCLE:
		_update_circle_item();
		break;
	case LINE:
		_update_line_item();
		break;
	case SELECT:
		_move_item();
		break;
	case CLEAN:
	case COLOR_SELECTOR:
	case REMOVE:
		break;
	default:
		dlog_print(DLOG_ERROR, LOG_TAG, "Unknown mode");
		break;
	}
}

/**
 * @brief Internal function which updates a freehand object by adjusting its position
 * according to current mouse pointer coordinates and applies the currently selected
 * pen color. If the object does not exist, then it is prior created.
 * The freehand object is drawn by concatenation of short lines connecting subsequent
 * points on mouse pointer movement path.
 */
static void _update_freehand_item(void)
{
	int x1 = 0;
	int y1 = 0;
	int x2 = 0;
	int y2 = 0;
	Evas_Object *part = NULL;

	if (!s_info.mouse_pressed)
		return;

	if (!s_info.current_object) {
		s_info.current_object = _create_object(FREEHAND);
		s_info.objects = eina_list_append(s_info.objects, s_info.current_object);
	}

	part = evas_object_line_add(s_info.evas);
	if (!part) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_line_add() failed. Cannot draw freehand.");
		return;
	}

	_add_object_part(s_info.current_object, part);

	if (s_info.prev.y <= s_info.curr.y) {
		x1 = s_info.prev.x;
		x2 = s_info.curr.x;
		y1 = s_info.prev.y;
		y2 = s_info.curr.y;
	} else {
		x2 = s_info.prev.x;
		x1 = s_info.curr.x;
		y2 = s_info.prev.y;
		y1 = s_info.curr.y;
	}

	evas_object_color_set(part, s_info.cl_pen.r, s_info.cl_pen.g, s_info.cl_pen.b, s_info.cl_pen.a);
	evas_object_line_xy_set(part, x1, y1, x2, y2);

	evas_object_repeat_events_set(part, EINA_TRUE);
	evas_object_show(part);
	evas_object_layer_set(part, s_info.win_layer + 1);
}

/**
 * @brief Internal function which updates a rectangle object by adjusting its position and size
 * according to current mouse pointer coordinates. If the object does not exist,
 * then it is prior created with currently selected pen and fill colors.
 * Each rectangle object consists of two rectangular primitives, where the second one
 * (drawn with fill color) is smaller than the first one (drawn with the pen color)
 * by doubled border size. Both rectangles are aligned to their centers.
 */
static void _update_rect_item(void)
{
	int x1 = 0;
	int y1 = 0;
	int x2 = 0;
	int y2 = 0;
	Evas_Object *part = NULL;

	if (!s_info.mouse_pressed)
		return;

	if (!s_info.current_object) {
		_create_rect();
		return;
	}

	if (s_info.start.x < s_info.curr.x) {
		x1 = s_info.start.x;
		x2 = s_info.curr.x - s_info.start.x;
	} else {
		x1 = s_info.curr.x;
		x2 = s_info.start.x - s_info.curr.x;
	}

	if (s_info.start.y < s_info.curr.y) {
		y1 = s_info.start.y;
		y2 = s_info.curr.y - s_info.start.y;
	} else {
		y1 = s_info.curr.y;
		y2 = s_info.start.y - s_info.curr.y;
	}

	part = eina_list_nth(s_info.current_object->parts, 0);
	evas_object_move(part, x1, y1);
	evas_object_resize(part, x2, y2);

	part = eina_list_nth(s_info.current_object->parts, 1);
	evas_object_move(part, x1 + OBJECT_BORDER, y1 + OBJECT_BORDER);
	evas_object_resize(part, x2 - (OBJECT_BORDER*2), y2 - (OBJECT_BORDER*2));
}

/**
 * @brief Internal function which updates a circle/ellipse object by adjusting its position and size
 * according to current mouse pointer coordinates. If the object does not exist,
 * then it is prior created with currently selected pen and fill colors.
 * Each circle/ellipse object consists of two circular/elliptic primitives, where the second one
 * (drawn with fill color) is smaller than the first one (drawn with the pen color)
 * by doubled border size. Both circles/ellipsis are aligned to their centers.
 */
static void _update_circle_item(void)
{
	Evas_Object *part = NULL;
	int radius_x = 0;
	int radius_y = 0;
	int center_x = 0;
	int center_y = 0;

	if (!s_info.mouse_pressed)
		return;

	if (!s_info.current_object) {
		s_info.current_object = _create_object(CIRCLE);
		_create_circle_part(s_info.current_object, s_info.cl_pen.r, s_info.cl_pen.g, s_info.cl_pen.b, s_info.cl_pen.a);
		_create_circle_part(s_info.current_object, s_info.cl_fill.r, s_info.cl_fill.g, s_info.cl_fill.b, s_info.cl_fill.a);
		s_info.objects = eina_list_append(s_info.objects, s_info.current_object);
	}

	radius_x = (s_info.curr.x - s_info.start.x) / 2;
	radius_y = (s_info.curr.y - s_info.start.y) / 2;

	center_x = s_info.start.x + radius_x;
	center_y = s_info.start.y + radius_y;

	radius_x = abs(radius_x);
	radius_y = abs(radius_y);

	part = eina_list_nth(s_info.current_object->parts, 0);
	_update_circle_part(part, radius_x, radius_y, center_x, center_y);

	part = eina_list_nth(s_info.current_object->parts, 1);
	_update_circle_part(part, radius_x - OBJECT_BORDER, radius_y - OBJECT_BORDER, center_x, center_y);
}

/**
 * @brief Internal function which updates a line object by adjusting its position and size
 * according to current mouse pointer coordinates. If the object does not exist,
 * then it is prior created with currently selected pen color.
 */
static void _update_line_item(void)
{
	int x1 = 0;
	int y1 = 0;
	int x2 = 0;
	int y2 = 0;

	if (!s_info.mouse_pressed)
		return;

	if (!s_info.current_object) {
		_create_line();
		return;
	}

	if (s_info.start.y <= s_info.curr.y) {
		x1 = s_info.start.x;
		x2 = s_info.curr.x;
		y1 = s_info.start.y;
		y2 = s_info.curr.y;
	} else {
		x2 = s_info.start.x;
		x1 = s_info.curr.x;
		y2 = s_info.start.y;
		y1 = s_info.curr.y;
	}

	evas_object_line_xy_set(_get_first_part_of_current_object(), x1, y1, x2, y2);
}

/**
 * @brief Internal function which performs selected object reposition according to
 * the mouse pointer movement path. This function keeps track of the reposition
 * coordinates taking into account target object's size in order to avoid moving
 * the object outside the drawing area.
 */
static void _move_item(void)
{
	Evas_Object *part = NULL;
	int x = 0;
	int y = 0;
	int offset_x = 0;
	int offset_y = 0;
	Eina_List *l = NULL;
	int draw_area_y;

	if (!s_info.mouse_pressed)
		return;

	if (!s_info.selected)
		return;

	/*
	 * If the color palette is visible then the drawing area is shrunk
	 * to the bottom edge of the color palette.
	 */
	draw_area_y = _get_draw_area_top();
	evas_object_geometry_get(s_info.selection_frame, &x, &y, NULL, NULL);

	offset_x = s_info.curr.x - s_info.prev.x;
	offset_y = s_info.curr.y - s_info.prev.y;

	/* Frame position is used, so there's no need to recalculate drawn objects geometry.
	 * Moving out of the draw area is checked below. */
	if (y + offset_y <= draw_area_y)
		return;

	evas_object_move(s_info.selection_frame, x + offset_x, y + offset_y);

	EINA_LIST_FOREACH(s_info.selected->parts, l, part) {
		if (!part)
			continue;

		evas_object_geometry_get(part, &x, &y, NULL, NULL);
		evas_object_move(part, x + offset_x, y + offset_y);
	}
}

/**
 * @brief Internal function which creates an abstraction of common object with
 * given type: FREEHAND, LINE, CIRCLE, RECTANGLE.
 * @param[in] type The type of the object to be created.
 * @return This function returns an object's abstraction if it was created successfully,
 * otherwise NULL is returned.
 */
static object_t *_create_object(mode_type_t type)
{
	object_t *object = (object_t *)calloc(1, sizeof(object_t));
	if (!object) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_create_object() failed. Memory allocation error.");
		return NULL;
	}

	object->type = type;
	object->parts = NULL;

	return object;
}

/**
 * @brief Internal function which adds a part to the object's abstraction.
 * The newly added part represents a drawn object's surface.
 * @param[in] object The object's abstraction.
 * @param[in] part The part to be added to the object's abstraction.
 */
static void _add_object_part(object_t *object, Evas_Object *part)
{
	if (!object || !part) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_add_object_part() failed. Invalid input argument.");
		return;
	}

	evas_object_data_set(part, OBJECT_DATA, object);
	object->parts = eina_list_append(object->parts, part);
	evas_object_event_callback_add(part, EVAS_CALLBACK_MOUSE_DOWN, _object_clicked_cb, NULL);
}

/**
 * @brief Internal function which creates a rectangle object.
 * The rectangle object is drawn by merging two rectangular parts and is stored in
 * object's abstraction data structure.
 */
static void _create_rect(void)
{
	s_info.current_object = _create_object(RECTANGLE);
	if (!s_info.current_object)
		return;

	_add_rect_part(s_info.current_object, s_info.cl_pen.r, s_info.cl_pen.g, s_info.cl_pen.b, s_info.cl_pen.a, s_info.start.x, s_info.start.y);
	_add_rect_part(s_info.current_object, s_info.cl_fill.r, s_info.cl_fill.g, s_info.cl_fill.b, s_info.cl_fill.a, s_info.start.x, s_info.start.y);

	s_info.objects = eina_list_append(s_info.objects, s_info.current_object);
}

/**
 * @brief Internal function which adds the rectangular part to the drawing canvas
 * with given color at provided coordinates. The part's size is not adjusted in this function.
 * For size update, check the _update_rect_item() function.
 * @param[in] object The object's abstraction.
 * @param[in] r The red component of the color to be applied.
 * @param[in] g The green component of the color to be applied.
 * @param[in] b The blue component of the color to be applied.
 * @param[in] a The alpha channel for the color to be applied.
 * @param[in] x The X coordinate of the upper-left corner of the part.
 * @param[in] y The Y coordinate of the upper-left corner of the part.
 * @return This function returns rectangular part if it was created successfully,
 * otherwise NULL is returned.
 */
static Evas_Object *_add_rect_part(object_t *object, int r, int g, int b, int a, int x, int y)
{
	Evas_Object *part = evas_object_rectangle_add(s_info.evas);
	if (!part) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_rectangle_add() failed.");
		return NULL;
	}

	evas_object_color_set(part, r, g, b, a);
	evas_object_move(part, x, y);

	evas_object_repeat_events_set(part, EINA_TRUE);
	evas_object_show(part);
	evas_object_layer_set(part, s_info.win_layer + 1);
	_add_object_part(object, part);

	return part;
}

/**
 * @brief Internal function which creates a line object.
 * The line object is drawn with currently selected pen color and is stored in
 * object's abstraction data structure.
 */
static void _create_line(void)
{
	Evas_Object *line_new = NULL;
	line_new = evas_object_line_add(s_info.evas);
	if (!line_new) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_line_add() failed.");
		return;
	}

	s_info.current_object = _create_object(LINE);
	if (!s_info.current_object)
		return;

	evas_object_color_set(line_new, s_info.cl_pen.r, s_info.cl_pen.g, s_info.cl_pen.b, s_info.cl_pen.a);
	evas_object_line_xy_set(line_new, s_info.start.x, s_info.start.y, s_info.start.x, s_info.start.y);

	evas_object_repeat_events_set(line_new, EINA_TRUE);
	evas_object_show(line_new);
	evas_object_layer_set(line_new, s_info.win_layer + 1);

	_add_object_part(s_info.current_object, line_new);
	s_info.objects = eina_list_append(s_info.objects, s_info.current_object);
}

/**
 * @brief Internal function which adds the polygonal part to the drawing canvas
 * with given color. The part's size and shape are not adjusted in this function.
 * For size update, check the _update_circle_item() function. For shape update,
 * check the _update_circle_part() function.
 * @param[in] object The object's abstraction.
 * @param[in] r The red component of the color to be applied.
 * @param[in] g The green component of the color to be applied.
 * @param[in] b The blue component of the color to be applied.
 * @param[in] a The alpha channel for the color to be applied.
 * @return This function returns polygonal part if it was created successfully,
 * otherwise NULL is returned.
 */
static Evas_Object *_create_circle_part(object_t *object, int r, int g, int b, int a)
{
	Evas_Object *part = NULL;

	if (!object) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_create_circle_part() failed. Invalid input argument.");
		return NULL;
	}

	part = evas_object_polygon_add(s_info.evas);
	if (!part) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_polygon_add() failed");
		return NULL;
	}

	_add_object_part(object, part);
	evas_object_color_set(part, r, g, b, a);
	evas_object_repeat_events_set(part, EINA_TRUE);
	evas_object_show(part);
	evas_object_layer_set(part, s_info.win_layer + 1);

	return part;
}

/**
 * @brief Internal function which reshapes the polygonal part to the form
 * of a circle/ellipse with the radius sizes initially set.
 * Internally, the polygon is reshaped by adding 360 anchor points shifted
 * by 1 degree around the center point.
 * @param[in] part The polygonal part to be updated.
 * @param[in] radius_x The radius along the X axis.
 * @param[in] radius_y The radius along the Y axis.
 * @param[in] center_x The X coordinate of the center point.
 * @param[in] center_y The Y coordinate of the center point.
 */
static void _update_circle_part(Evas_Object *part, int radius_x, int radius_y, int center_x, int center_y)
{
	int i;

	if (!part) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_update_circle_part() failed. Invalid input argument.");
		return;
	}

	evas_object_polygon_points_clear(part);

	for (i = 0; i < 360; ++i)
		evas_object_polygon_point_add(part, sin(i * M_PI / 180) * radius_x + center_x, cos(i * M_PI / 180) * radius_y + center_y);
}

/**
 * @brief Internal function which gets the extreme (edge) points of the object
 * which are defining the smallest circumrectangle for given object.
 * @param[in] item The object's abstraction to be analyzed.
 * @param[out] min_x The X coordinate of the smallest circumrectangle's left border.
 * @param[out] min_y The Y coordinate of the smallest circumrectangle's top border.
 * @param[out] max_x The X coordinate of the smallest circumrectangle's right border.
 * @param[out] max_y The Y coordinate of the smallest circumrectangle's bottom border.
 */
static void _get_item_geometry(object_t *item, int *min_x, int *min_y, int *max_x, int *max_y)
{
	int x = 0;
	int y = 0;
	int w = 0;
	int h = 0;
	Eina_List *l = NULL;
	Evas_Object *part;

	if (!item || !item->parts) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_get_item_geometry() failed. Invalid input argument.");
		return;
	}

	part = eina_list_data_get(item->parts);
	evas_object_geometry_get(part, min_x, min_y, max_x, max_y);
	*max_x = *min_x + *max_x;
	*max_y = *min_y + *max_y;

	EINA_LIST_FOREACH(item->parts, l, part) {
		evas_object_geometry_get(part, &x, &y, &w, &h);

		if (x < *min_x)
			*min_x = x;

		if (y < *min_y)
			*min_y = y;

		if (x + w > *max_x)
			*max_x = x + w;

		if (y + h > *max_y)
			*max_y = y + h;
	}
}

/**
 * @brief Internal function which gets the first part of currently selected object.
 * @return This function returns the first part's Evas object if it was successfully obtained,
 * otherwise NULL is returned.
 */
static Evas_Object *_get_first_part_of_current_object(void)
{
	if (!s_info.current_object) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_get_first_part_of_current_object() failed.");
		return NULL;
	}

	return eina_list_data_get(s_info.current_object->parts);
}

/**
 * @brief Internal function which checks whether the cursor coordinates are
 * contained inside the selection frame.
 * @return This function returns EINA_TRUE if the cursor is inside the selection
 * frame, otherwise 'false' is returned.
 */
static Eina_Bool _is_cursor_in_selection_frame(void)
{
	int x = 0;
	int y = 0;
	int w = 0;
	int h = 0;

	if (!s_info.selection_frame)
		return EINA_FALSE;

	evas_object_geometry_get(s_info.selection_frame, &x, &y, &w, &h);

	return (s_info.start.x >= x && s_info.start.y >= y && s_info.start.x <= x + w && s_info.start.y <= y + h && evas_object_visible_get(s_info.selection_frame));
}

/**
 * @brief Internal function which sets the operational mode.
 * @param[in] mode The operational mode to be set.
 * Available modes:
 * - CLEAN - cleans the drawing area;
 * - FREEHAND - freehand drawing mode;
 * - COLOR_SELECTOR - color selection mode;
 * - LINE - line drawing mode;
 * - CIRCLE - circle/ellipse drawing mode;
 * - RECTANGLE - rectangle drawing mode;
 * - SELECT - object's selection mode;
 * - REMOVE - objects removal mode.
 */
static void _set_toolbar_mode(mode_type_t mode)
{
	s_info.mode = mode;
}

/**
 * @brief Internal function which check whether two strings are the same.
 * @param[in] s1 First string.
 * @param[in] s2 Second string.
 * @return This function returns 'true' if both strings are the same,
 * otherwise 'false' is returned.
 */
static bool _is_same_string(const char *s1, const char *s2)
{
	if (!s1 || !s2)
		return false;

	return (strncmp(s1, s2, strlen(s1)) == 0 && strlen(s1) == strlen(s2));
}
